'''
Class which returns markers for a given 
qualisys file based on catagory.
'''

from re import compile

# Set up the regular expressions that will allow us to go up the leg ticking off joints and bones
reAcc = compile("[FB][RL]Acc")
reGyro = compile("[FB][RL]Gyro")
reMM1Hoof=compile('hoof_dorsal')
reMM2Hoof=compile('hoof_lateral')
reDigit=compile('([a-zA-Z0-9]*(Digit)|(Toe)[a-zA-Z0-9]*)(?i)') # The 'toe' is best thought of as the end of the chain of bones from hip down
reDIP=compile('([a-zA-Z0-9]*(Dip)[a-zA-Z0-9]*)(?i)')
#rePhalanges=compile('([a-zA-Z0-9]*(Phalanges)[a-zA-Z0-9]*)(?i)')
reMTP=compile('([a-zA-Z0-9]*(Fetlock)|(EDL)|(MTP)|(MCP)|(TMP)[a-zA-Z0-9]*)(?i)')
reMetatarsus=compile('([a-zA-Z0-9]*(Metatarsus)[a-zA-Z0-9]*)|([a-zA-Z0-9]*(Foot)[a-zA-Z0-9]*)(?i)') # Metatarsus is on the foot so put 'foot' labels here.
reTMT=compile('([a-zA-Z0-9]*(TMT)[a-zA-Z0-9]*)(?i)')
reTarsus=compile('([a-zA-Z0-9]*(Tarsus)|(Carpal)|(Tarsal)|(Carpus)[a-zA-Z0-9]*)(?i)') # Heel is roughly the same as the tarsus so put it here
reHock=compile('([a-zA-Z0-9]*(Hock)[a-zA-Z0-9]*)|([a-zA-Z0-9]*(Calcaneus)[a-zA-Z0-9]*)|([a-zA-Z0-9]*(Heel)[a-zA-Z0-9]*)(?i)') # Heel is the Hock so put this here
reTibFib=compile('([a-zA-Z0-9]*(Tibia)[a-zA-Z0-9]*)|([a-zA-Z0-9]*(Fibia)[a-zA-Z0-9]*)(?i)')
reStifel=compile('([a-zA-Z0-9]*(Stifle)[a-zA-Z0-9]*)|([a-zA-Z0-9]*(Patella)[a-zA-Z0-9]*)|([a-zA-Z0-9]*(Knee)[a-zA-Z0-9]*)|([a-zA-Z0-9]*(Elbow)[a-zA-Z0-9]*)(?i)') # Where the knee roughly is
reFemur=compile('([a-zA-Z0-9]*(Femur)[a-zA-Z0-9]*)(?i)')
reHip=compile('([a-zA-Z0-9]*(Hip)|(SHO)[a-zA-Z0-9]*)(?i)')
reDistalScapula=compile('([a-zA-Z0-9]*(Disscap)[a-zA-Z0-9]*)(?i)')
reScapula=compile('([a-zA-Z0-9]*(ScapulaRight)|(ScapulaLeft)|(Scapula(!?((_ventral)|(_dorsal))))[a-zA-Z0-9]*)(?i)')
rePelivs=compile('([a-zA-Z0-9]*(Pelvis)|(Proscap)|(Shoudler)|(Shoulder)|(tc)[a-zA-Z0-9]*)(?i)')
legBoneAnatomy=[reAcc, reGyro, reMM1Hoof,reMM2Hoof,reDigit,reDIP,reMTP,reMetatarsus,reTarsus,reHock,reTibFib,reStifel,reFemur,reHip,reScapula,reDistalScapula,rePelivs]
reHoof=compile('(^Toe[a-zA-Z0-9]*)|(^toe[a-zA-Z0-9]*)|(^heel[a-zA-Z0-9]*)|(^dip[a-zA-Z0-9]*)' )

# Now list all markers that will be labelled as head
reHead=compile('([a-zA-Z0-9]*(Head)[a-zA-Z0-9]*)|([a-zA-Z0-9]*(Eye)[a-zA-Z0-9]*)|([a-zA-Z0-9]*(?<!(R|r))(Ear)[a-zA-Z0-9]*)|([a-zA-Z0-9]*(Nose)[a-zA-Z0-9]*)|([a-zA-Z0-9]*(Poll)[a-zA-Z0-9]*)(?i)')

# Now list all body markers, not that tail base is a body marker
reBody=compile('([a-zA-Z0-9]*(iliacCrest)|(rtc)|(ltc)|(sacrum)|(Back)|(Belly)|(Body)|(TailBase)|(Spine)|(Thorax)|(Pelvis)|(Illiac)|(Wither(s?))|(Lumbar)|(Cranial)|(Caudal)[a-zA-Z0-9]*)(?i)')

# Now list all tail markers, should not include anythin labelled 'tailbase' so 
# go a bit conservative here
reTail=compile('([a-zA-Z0-9]*(TailTip)[a-zA-Z0-9]*)')

# Define fore and aft and left and right
reLeft=compile('[FB]LAcc|([a-zA-Z0-9]*(Left)|(left)[a-zA-Z0-9]*)|(l_scapula_dorsal)|(l_scapula_ventral)|(l[fh]_[_a-zA-Z0-9]*)|(L[FH]_[_a-zA-Z0-9]*)|([a-z]+L[_a-zA-Z0-9]*)')
reRight=compile('[FB]RAcc|([a-zA-Z0-9]*(Right)|(right)[a-zA-Z0-9]*)|(r_scapula_dorsal)|(r_scapula_ventral)|(r[fh]_[_a-zA-Z0-9]*)|([a-z]+R(?!ear)|(R[FH]_[_a-zA-Z0-9]*)[_a-zA-Z0-9]*)')
reFore=compile('[F][RL]Acc|([a-zA-Z0-9]*(Fore)|(shoudler)|(shoulder)|(Elbow)|(MCP)|(r_scapula_dorsal)|(r_scapula_ventral)|(l_scapula_dorsal)|(l_scapula_ventral)|(Shoulder)|(Scapula)|(scapula)|(carpal)|(Front)[a-zA-Z0-9]*)|(^[a-z]+[RL]F[_a-zA-Z0-9]*$)|(^mcp[LR][_a-zA-Z0-9]*)|(^[rl]f_[a-z]+)|(^[RL]F_[a-z]+)|(^discarpus[LR][_a-zA-Z0-9]*)|(^procarpus[LR][_a-zA-Z0-9]*)|(^elbow[LR][_a-zA-Z0-9]*)|(^sho[LR][_a-zA-Z0-9]*)|(^disscap[LR][_a-zA-Z0-9]*)|(^proscap[LR][_a-zA-Z0-9]*)|([a-zA-Z0-9]*(Carpus)|(Carpal)[a-zA-Z0-9]*)')
reHind=compile('[B][RL]Acc|([a-zA-Z0-9]*(Rear)|(Hind)|(pelvisRight)|(Stifle)|(pelvisLeft)|(MTP)|(RightPelvis)|(LeftPelvis)|(Hip)[a-zA-Z0-9]*)|(^[a-z]+[RL]H)|(^[rl]h_[a-z]+)|(^[RL]H_[a-z]+)|(^mtp[LR])|(^distarsus[LR])|(^protarsus[LR])|(^stifle[LR])|(^hip[LR])|(^tc[LR])|([a-zA-Z0-9]*(Tarsus)[a-zA-Z0-9]*)|([a-zA-Z0-9]*(tarsus)|(tarsal)[a-zA-Z0-9]*)')

# Define markers that are labelled but have nothing to do with an animal or robot
reJunkMarkers=compile('([a-zA-Z0-9]*(Camera)|(Obstacle)[a-zA-Z0-9]*)|(marker[0-9])(?i)')

class ToeID(object):
    '''
    Encapsulates information about where a toe is on a body in general terms.
    This allows us to work out what other markers are on that appendage.
    '''
    def __init__(self,name,legConditions):
        self.name=name
        self.legConditions=legConditions

    def sameLeg(self,marker):
        isSame=True
        for c in self.legConditions:   # Check each condition to be in this leg
            if c.findall(marker)==[]:  # If we are in the same leg this must not be empty
                isSame=False
        return(isSame)

def identifyQuadrupedToe(toe):
    # Check if the toe is on the left, then check if it is fore or aft
    cond=[]
    id=""
    if reFore.findall(toe)!=[]:
        id+="Fore"
        cond.append(reFore)
    elif reHind.findall(toe)!=[]:
        id+="Hind"
        cond.append(reHind)
    else:
        raise(Exception("In a quadruped a foot marker must either be hind or fore!"))
    if reLeft.findall(toe)!=[]:
        id+="Left"
        cond.append(reLeft)
    elif reRight.findall(toe)!=[]:
        id+="Right"
        cond.append(reRight)
    else:
        raise(Exception("In a quadruped a foot marker must either be left or right!"))
    return(ToeID(id,cond))

class Markers(object):
  def __init__(self,markers):
    self.markers = markers
    self.headMarkers=[k for k in self.markers if reHead.findall(k)!=[]]
    self.bodyMarkers=[k for k in self.markers if reBody.findall(k)!=[]]
    self.junkMarkers=[k for k in self.markers if reJunkMarkers.findall(k)!=[]]
    self.tailMarkers=[k for k in self.markers if reTail.findall(k)!=[]]
    self.appendages={}
    for p in legBoneAnatomy:
        toes=[m for m in self.markers if p.findall(m)!=[]] 
        if toes!=[]:
            break
    self.numberOfToes=len(toes)
    for t in toes:
        toeID=identifyQuadrupedToe(t) # Call the method for n-1 legs
        legCohabitants=[m for m in markers if toeID.sameLeg(m)==True]
        self.appendages[toeID.name]=[]
        for p in legBoneAnatomy:
            for m in legCohabitants:
                if p.findall(m)!=[] and [foundMarker for foundMarker in self.appendages[toeID.name] if foundMarker==m] == []:
                    self.appendages[toeID.name].append(m)
