'''
Compare Schoner to general model
'''

# Options
from options import glbOptions,getDirs,dogs,morph

from numpy import (array,unwrap,vstack,hstack,sin,cos,sum,min,pi,dot,std,
  arange)
from scipy.stats import f as fstat
from scipy.stats import skew
from scipy.interpolate import UnivariateSpline
from numpy import random,triu_indices,ones,log,nan,newaxis,linspace
from numpy.linalg import lstsq
from pylab import (figure,plot,subplot,hist,xlabel,ylabel,rc,xticks,boxplot,
  savefig,scatter,tight_layout,fill_between,xlim,legend)

from pickle import load
from os.path import join as osjoin

font = {'family' : 'normal',
        'size'   : 20}

rc('font', **font)

random.seed(0) #Force consistent results

sr = 177.0

markers = ['x','+','*','D','o']
colors = ['r','b','c','m','g']

# Group elements
G = ["e","lr","tfh","tfhlr"]
# Model order
order = 2

def AInfCrit(k,n,RSS):
  return(2*k+n*log(RSS/n))

# Results
M = dict()
K = dict()
dP = dict()
res = dict()
rss = dict()
AIC = dict()
F = dict()
Ftest = dict()
Skew = dict()

for dog in dogs:
  print "Doing dog:", dog
  glb = glbOptions(dog=dog)

  dataDir,baseDir,outDir,intDir,logDir,plotDir = getDirs(glb.dog,glb.dataDir)

  print "Loading data"
  fn = osjoin(outDir,"transitions.pkl")
  f = open(fn,'r')
  gTType,sigs,phases,theta,testBeltV,testvdog,testsvdog,phi = load(f)
  f.close()

  # Get all Walk Trot and Trot Walk transitions
  print "Assembling walk and trot transitions"
  trans = [k for k in gTType if gTType[k]=="WalkTrot" or gTType[k]=="TrotWalk"]
  tpe = array([
    "WalkTrot" if gTType[k]=="WalkTrot" else "TrotWalk" 
    for k in gTType if gTType[k]=="WalkTrot" or gTType[k]=="TrotWalk"
  ])
  wTTheta = dict([(k,theta[k]) for k in trans])
  allv = [
    testsvdog[k]
    for k in trans
    if gTType[k]=="WalkTrot" or gTType[k]=="TrotWalk"
  ]

  # Form into arrays
  legs = phases[phases.keys()[0]].keys()
  p = [array([unwrap(phases[k][l]) for l in legs]) for k in trans]
  dp = [array([
    UnivariateSpline(arange(q.size),q,s=0).derivative()(arange(q.size))
    for q in ip
  ]) for ip in p]
  p = hstack([ip[:,:] for ip in p])
  dp = hstack(dp)
  v = hstack([iv[:] for iv in allv])

  pd = array([p[i]-p[j] for i,j in zip(*triu_indices(4,1))])
  s = vstack([sin(n*pd) for n in xrange(1,order+1)])
  c = vstack([cos(n*pd) for n in xrange(1,order+1)])
  a = ones((1,p.shape[1]))
  m1 = vstack([a,s,c,a*v,s*v,c*v]).T
  mt = vstack([-a,s,-c,-a*v,s*v,-c*v]).T
  
  print "Fitting Model"
  M[dog] = dict()
  K[dog] = dict()
  dP[dog] = dict()
  res[dog] = dict()
  rss[dog] = dict()
  AIC[dog] = dict()
  F[dog] = dict()
  Ftest[dog] = dict()
  Skew[dog] = dict()
  dP[dog]["tfhlr"] = hstack(dp)[:,newaxis]
  M[dog]["tfhlr"] = vstack([m1,mt,mt,m1])
  dP[dog]["e"] = dp.T
  M[dog]["e"] = m1.copy()
  dP[dog]["lr"] = array([hstack(dp[[0,3]]),hstack(dp[[1,2]])]).T
  M[dog]["lr"] = vstack([m1,m1])
  dP[dog]["tfh"] = array([hstack(dp[[0,2]]),hstack(dp[[1,3]])]).T
  M[dog]["tfh"] = vstack([m1,mt])
  
  for g in reversed(G):
    K[dog][g] = lstsq(M[dog][g],dP[dog][g])[0]
    res[dog][g] = (dot(M[dog][g],K[dog][g])-dP[dog][g])
    rss[dog][g] = sum(res[dog][g]**2)
    AIC[dog][g] = AInfCrit(K[dog][g].size,res[dog][g].size,rss[dog][g])
    F[dog][g] = nan if g=="tfhlr" else (
        (rss[dog]["tfhlr"]-rss[dog][g])/float(K[dog][g].size-K[dog]["tfhlr"].size)
      ) / (
        rss[dog][g]/float(dp.size-K[dog][g].size)
      )
    Ftest[dog][g] = nan if g=="tfhlr" else 1.0-fstat.cdf(
      F[dog][g],K[dog][g].size-K[dog]["tfhlr"].size,dp.size-K[dog][g].size
    )
    Skew[dog][g] = skew(res[dog][g].flatten())
    
print "Done all basic models"

'''  
# Do a step AIC version
V = [M[d]['e'] for d in dogs]
U = [dP[d]['e'] for d in dogs]
k = range(V[0].shape[1])
nk = array(k)

S = array([lstsq(v,u)[0] for v,u in zip(V,U)])
R = [(dot(v,s)-u) for v,u,s in zip(V,U,S)]
r2 = sum([sum(r**2) for r in R])
AIC0 = AInfCrit(S.size,sum([r.size for r in R]),r2)
scores = dict()
scores[tuple(k)] = {"rss":float(r2),"aic":float(AIC0)}
nAIC=float(AIC0)
AIC0 = nAIC+1.0
while(nAIC<AIC0):
  AIC0 = float(nAIC)
  k = array(nk)
  print "Current AIC:",AIC0
  print "Current k:",k
  for i in xrange(len(k)):
    kt = concatenate((k[:i],k[i+1:]))
    V2 = [v[:,kt] for v in V]
    S = array([lstsq(v,u)[0] for v,u in zip(V2,U)])
    R = [(dot(v,s)-u) for v,u,s in zip(V2,U,S)]
    r2 = sum([sum(r**2) for r in R])
    ic = AInfCrit(S.size,sum([r.size for r in R]),r2)
    if ic<nAIC:
      nAIC=float(ic)
      nk = array(kt)
      nr2 = float(r2)
    print ".",
  print ""
  if nAIC!=AIC0:
    scores[tuple(k)] = {"rss":float(nr2),"aic":float(nAIC)}
'''

from scipy.optimize import fmin

print "Phase shifted model fit"
# Now try to minimize symmetry violation via a phase shift
def phaseOptim(delta,V,U,full=False):
  U = U.copy() + delta[0]*array([0,1,1,0])
  U = hstack(U.T)[:,newaxis]
  S = lstsq(V,U)[0]
  R = dot(V,S)-U
  R2 = sum(R**2)
  if full:
    return(S,R,R2,AInfCrit(S.size,R.size,R2),skew(R.flatten()))
  print delta,R2
  return(R2)

delta = dict()
deltaScore = dict()
for d in dogs:
  print "Doing dog:",d
  deltaScore[d] = None
  delta[d] = 0.0
  for tau in linspace(-pi/5,pi/5,50):
    nScore = phaseOptim([tau],M[d]['tfhlr'],dP[d]['e'])
    if (deltaScore[d]==None or nScore<deltaScore[d]):
      deltaScore[d] = nScore
      delta[d] = tau
  delta[d] = fmin(phaseOptim,[delta[d]],args=(M[d]['tfhlr'],dP[d]['e']),xtol=1e-7,ftol=1e-9)

for d in dogs:
  print d,"phase shift:",delta[d][0],"AIC:",phaseOptim([delta[d]],M[d]['tfhlr'],dP[d]['e'],True)[-2],"Skew:",phaseOptim([delta[d]],M[d]['tfhlr'],dP[d]['e'],True)[-1]

for d in dogs:
  print d,"RSS:",phaseOptim([delta[d]],M[d]['tfhlr'],dP[d]['e'],True)[-3],"phase shift:",delta[d][0]
  
from sipylibs import linearRegression
x = array([(delta[d][0],m) for d,m in zip(dogs,morph)]).T
r = linearRegression(*x[[1,0]])
morphs = linspace(min(x[1])-std(x[1])/10.0,max(x[1])+std(x[1])/10.0,1000)

##############################################################################
# RESULTS
##############################################################################
for dog in dogs:
  print dog
  print "RSS:"
  print rss[dog]
  print "AIC:"
  print AIC[dog]
  print "F"
  print F[dog]
  print "p"
  print Ftest[dog]
  print "S"
  print Skew[dog]
  figure()
  for g in G:
    hist(hstack(res[dog][g]),100,alpha=0.3,label=g)
  legend()

# Mass vs. symmetry violation
figure(figsize=(8,16))
subplot(2,1,2)
for d,m,c,l in zip(dogs,morph,colors,markers):
  scatter(m,delta[d][0],color=c,marker=l,s=36,lw=2,alpha=0.7)
plot(morphs,r.predict(morphs),c='k',lw=2)
plot(morphs,r.confidenceBounds(morphs)[0],c='b',lw=2)
plot(morphs,r.confidenceBounds(morphs)[1],c='b',lw=2)
fill_between(morphs,r.confidenceBounds(morphs)[0],r.confidenceBounds(morphs)[1],alpha=0.1,color='b')
xlabel("aspect ratio")
ylabel("phase offset (rad)")
xlim((morphs.min(),morphs.max()))
tight_layout()
savefig(osjoin(glb.basePath,"massVsAsymetryTerm.png"))
savefig(osjoin(glb.basePath,"massVsAsymetryTerm.svg"))
print "Is phase shift necessary to restore symmetry related to morphology?"
print r.testSlope()

perfMat = hstack([
  [[rss[d][g] for g in G] for d in dogs],
  array([[phaseOptim([delta[d]],M[d]['tfhlr'],dP[d]['e'],True)[-3] for d in dogs]]).T
])

ax = subplot(2,1,1)
ax.semilogy()
boxplot(perfMat)
for i,p in enumerate(perfMat.T):
  for ip,c,l in zip(p,colors,markers):
    scatter(i+1.4,ip,marker=l,color=c,lw=3,alpha=0.7)
xticks(arange(5)+1,["$\mathbf{C}_1$","$\mathbf{C}_{lr}$","$\mathbf{C}_{fh}$","$\mathbf{C}$","$\mathbf{C}$ shifted"])
ylabel("residual sum of squared ($rad^2/s^2$)")
tight_layout()
savefig(osjoin(glb.basePath,"symmetryViolationEstimate.png"))
savefig(osjoin(glb.basePath,"symmetryViolationEstimate.svg"))

# Check everything consistent with leave one out validation
gamma = 0.95
figure()
for d,m,c,l in zip(dogs,morph,colors,markers):
  scatter(m,delta[d][0],color=c,marker=l,s=36,lw=2,alpha=0.7)
for i,c in enumerate(colors):
  xp = array([(delta[d][0],m) for j,(d,m) in enumerate(zip(dogs,morph)) if i!=j]).T
  rp = linearRegression(*xp[[1,0]])
  plot(morphs,rp.predict(morphs),c=c,lw=4)  
  plot(morphs,rp.confidenceBounds(morphs,gamma)[0],c=c,lw=2)
  plot(morphs,rp.confidenceBounds(morphs,gamma)[1],c=c,lw=2)
  fill_between(
    morphs,rp.confidenceBounds(morphs,gamma)[0],
    rp.confidenceBounds(morphs,gamma)[1],
    alpha=0.1,color=c
  )
  print rp.testSlope()
  print rp.beta
