'''
Compute phases from the gait transitions data set using different methods 
then plot residuals for some trials as a function of time.

@author Simon Wilshin
@email swilshin@rvc.ac.uk
@date Jul 2013
'''

from markers import Markers
from directories import getDirs
from utility import splineDiff,tukeyBiweight,dtukeyBiweight
from options import glbOptions

from sipylibs.qualisys import QualisysRawReader
from sipylibs import FiltFiltFilter,zscore
from sipylibs.interpolation import nanInterpolation
from sipylibs import PCA

from phaserng.dforms import ( CompositeRandomBasisFitter,ThetaFitter,
  FourierFitter,TangentVectors)
from phaserng import PhaserNG

from phaserng.phaserngutil import eventPhase,Phaser,FourierSeries

from numpy import (array,vstack,pi,linspace,unwrap,hstack,abs,mean,
  logical_and,sqrt,sum,std,dot,percentile,random,arange,polyfit,where,diff,
  polyval)
from numpy.linalg import inv
from pylab import (figure,plot,xlabel,ylabel,savefig,subplot,semilogy,
  xlim,ylim,hist,fill_between,fill_betweenx,vlines,Rectangle,axis)

from os.path import isfile
from os.path import join as osjoin
from os import listdir
from pickle import dump
from pickle import load
from sys import argv

# Plot mean square error
random.seed(0) # Make the random seed fixed so we can replicate exact results

phaseNames = ["phaser","form",'event']

cols = {
  'phaser':'k',
  'form':'#106AA4', #Teal
  'event':'#43BF3C', # Green
  'true':'#FF7F00' # Orange
}

markers = {
  'phaser':'s',
  'form':'D',
  'event':'h',
  'true':'+'
}

if '_dog' in locals():
  glb = glbOptions(dog=_dog)
elif len(argv)>=2 and argv[-1]!="--pylab":
  glb = glbOptions(dog=argv[-1])
else:
  print "WARNING: No dog specified, using default!"
  glb = glbOptions(None)

# Get files
print "Getting list of training files."
glb.dataDir,baseDir,outDir,intDir,logDir,plotDir = getDirs(glb.dog,glb.dataDir)
files = [f for f in listdir(intDir) if isfile(osjoin(intDir,f)) and f[-4:]==".tsv"]
trainFiles = [f for f in files if glb.reTraining.findall(f)!=[]]
trainType = ["Trot" if glb.reTraining.findall(f)[0]=="Trot" else "Walk" for f in trainFiles]

print "Loading training data" # Just get the walks
Qs = [QualisysRawReader(osjoin(intDir,fle)) for fle,tpe in zip(trainFiles,trainType) if tpe=="Walk"]

# Grab the training data
trainfle = osjoin(outDir,"trainingData.pkl")
T,phaserTrain,beltSpeed,vdog,svdog,trainType = load(open(trainfle,'r'))

T.cut = 0 # Stop cutting start and end off

if glb.dog=='Zack':
  glb.cpoints = [0,2]
else:
  glb.cpoints = [0,2]

lpf = FiltFiltFilter(Qs[0].FREQUENCY,50.0,4,'lowpass')

x0 = list()
y0 = list()
dy0 = list()
for Q in Qs:
  m = Markers(Q)
  figure()
  x0.append([])
  #y0.append(T(Q)["ForeLeft"].T)
  #dy0.append(splineDiff(y0[-1].T).T)
  for q in Q[m.appendages["ForeLeft"]].transpose(1,0,2):
    x0[-1].append((q[:,[0,2]]-T.getCoM(Q)[:,[0,2]]).T)
  #  subplot(1,2,1)
  #  plot(*x0[-1][-1])
  #subplot(1,2,2)
  #plot(*y0[-1])
  #xlim([-3,3])
  #ylim([-3,3])
  y0.append([])
  for m0,m1 in zip(m.appendages['ForeLeft'][:-1],m.appendages['ForeLeft'][1:]):
    p = (Q[m0][:,glb.cpoints]-Q[m1][:,glb.cpoints]).T
    y0[-1].append(p)
  y0[-1] = vstack(y0[-1])
  y0[-1] = vstack([lpf(y0[-1]),splineDiff(lpf(y0[-1]).T).T])
  plot(y0[-1].T)
y0=[vstack(ix) for ix in x0]
sy0 = list(y0)

# Do PCA
# Get Zscore constants and zscore
zskmu,zsksig,junk = zscore(hstack(y0),True)
y0 = [((iy.T-zskmu)/zsksig).T for iy in y0]
trs = hstack(y0).T
D,U = PCA(trs)
y0 = [dot(U.T,iy) for iy in y0]
pFac = 25
zmed = percentile(hstack(y0),pFac,1)+percentile(hstack(y0),100-pFac,1)
y0 = [(iy.T-zmed).T for iy in y0]
dy0 = [splineDiff(iy) for iy in y0]


ordP = 25
NpcF = 10
NpcP = 1

N = -5
yt = y0[:int(N)]
dyt = dy0[:int(N)]
yd = y0[int(N):]
xt = x0[:int(N)]
xd = x0[int(N):]
cxt = [array(ix).transpose(2,0,1) for ix in xt]
cxd = [array(ix).transpose(2,0,1) for ix in xd]
taut = [Q.Time for Q in Qs][:int(N)]
taud = [Q.Time for Q in Qs][int(N):]

'''
phi = phaserNG(
  [iyt[:NpcF] for iyt in yt],
  [idyt[:NpcF] for idyt in dyt],
  minR=0.05,ordPLim=ordP,fourCor2=True
)
'''

tt = TangentVectors(
  [iyt[:NpcF] for iyt in yt],
  [idyt[:NpcF] for idyt in dyt]
)

phi = PhaserNG(fourCor2=True,phi0=0.0)
phi.train(
  tt,
  ThetaFitter(minR=0.05),
  FourierFitter(ordPLim=ordP),
  CompositeRandomBasisFitter(pen=0.0,Nb=1000,lda=1.0)
)


psi = Phaser([iyt[:NpcP] for iyt in yt],ordP=ordP,psecfunc=lambda x: x[0])

pht = dict()
pht["phaser"] = [psi(iy[:NpcP])[0] for iy in yt]
pht["form"] = [phi(iy[:NpcF]) for iy in yt]
pht["event"] = [eventPhase(iy) for iy in yt]
phd = dict()
phd["phaser"] = [psi(iy[:NpcP])[0] for iy in yd]
phd["form"] = [phi(iy[:NpcF]) for iy in yd]
phd["event"] = [eventPhase(iy) for iy in yd]
for p,q in zip(pht["phaser"],pht["form"]):
  figure()
  plot((p-q)%(2*pi),c='r')

# Estimate limit cycle
theta = linspace(0,2*pi,251)
ls = dict()
fls = dict()
figure()
for k in phaseNames:
  ls[k] = list()
  cp = hstack(pht[k])
  for th0,th1 in zip(theta[:-1],theta[1:]):
    idx = logical_and(cp%(2*pi)>=th0,cp%(2*pi)<th1)
    ls[k].append(mean(vstack(cxt)[idx,:,:],0))
  ls[k] = array(ls[k])
  # If we are missing entries interpolate
  ls[k] = nanInterpolation(ls[k].reshape(ls[k].shape[0],-1).T,method="linear").T.reshape(ls[k].shape)
  #for ix in ls[k].transpose(1,0,2):
  #  plot(*ix.T)
  fls[k] = FourierSeries().fit(20,cp,vstack(cxt).reshape(cp.shape[0],-1).T) #FourierSeries().fit(phi.form.numParam,theta[:-1],ls[k].reshape(ls[k].shape[0],-1).T)
  for ix in fls[k].val(theta[:-1]).real.reshape(-1,5,2).transpose(1,0,2):
    plot(*ix.T,color=cols[k])

inx = dict()
res = dict()
figure()
for k in phd:
  inx[k] = [fls[k].val(p).real for p in phd[k]]
  res[k] = [sqrt(sum((ix-vstack(tx.transpose(1,2,0)).T)**2,1)) for ix,tx in zip(inx[k],cxd)]
  for r in res[k]:
    plot(r,c=cols[k])

errD = dict()
figure()
errD['event'],bins,dump = hist(hstack(res['event']),25,color=cols['event'],alpha=0.4,normed=True)
for k in ['phaser','form']:
   errD[k],bins,dump = hist(hstack(res[k]),bins,color=cols[k],alpha=0.4,normed=True)

figure(figsize=(21,9))
subplot(1,3,1)
for k in fls:
  fls[k] = FourierSeries().fit(ordP,theta[:-1],ls[k].reshape(ls[k].shape[0],-1).T)
  for ix in fls[k].val(theta[:-1]).real.reshape(-1,5,2).transpose(1,0,2):
    plot(*ix.T,c=cols[k],alpha=0.5,lw=2)   
xlabel("fore-aft position (mm)")
ylabel("vertical position (mm)")
subplot(1,3,2)
for k in phd:
  for r,t in zip(res[k],taud):
    plot(t-t[0],r,c=cols[k],lw=2,alpha=0.9)
xlabel("time (s)")
ylabel("root mean square error (mm)")
subplot(1,3,3)
for k in errD:
  plot(bins[:-1],errD[k],c=cols[k],marker=markers[k],lw=2)
semilogy(True)
xlabel("root mean square error (mm)")
ylabel("frequency density (1/mm)")
xlim([0,150])
ylim([0,max([max(errD[k]) for k in errD])*2.0])
savefig(glb.dog+"doglimbmodelperformance.png")


figure(figsize=(14,9))
subplot(1,2,1)
for k in fls:
  fls[k] = FourierSeries().fit(ordP,theta[:-1],ls[k].reshape(ls[k].shape[0],-1).T)
  for ix in fls[k].val(theta[:-1]).real.reshape(-1,5,2).transpose(1,0,2):
    plot(*ix.T,c=cols[k],alpha=0.5,lw=2)   
xlabel("fore-aft position (mm)")
ylabel("vertical position (mm)")
subplot(1,2,2)
for k in errD:
  plot(bins[:-1],errD[k],c=cols[k],marker=markers[k],lw=2)
semilogy(True)
xlabel("root mean square error (mm)")
ylabel("frequency density (1/mm)")
xlim([0,150])
ylim([0,max([max(errD[k]) for k in errD])*2.0])
savefig(glb.dog+"doglimbmodelperformancesub.png")


for k in res:
  print k,std(hstack(res[k])),hstack(res[k]).shape

window = 20
tau = 1.5
tauPad = 0.5
sr = Qs[0].FREQUENCY

perf = dict()
for k in phaseNames:
  perf[k] = list()

fdtme = list()

trl = 0
p = phd['event'][trl]
Ncyc = (unwrap(p)[-1]-unwrap(p)[0])/(2*pi)
Ns = len(p)
spad = (tau+tauPad)*Ns/Ncyc

figure()
for trl in xrange(-1*N):
  p = unwrap(phd['event'][trl])
  steps = where(diff(p[:-window-spad]%(2*pi))<-pi)[0]
  for idx,nidx in zip(steps[1:-1:2],steps[2::2]): #idx = random.randint(window+1,int(Ns-spad-window),1)[0]
    if len(p)>idx+window+(tau*Ns/Ncyc):
      for k in phaseNames:
        p = phd[k][trl]
        x = cxd[trl]
        
        # Do a standard polynomial fit to get sense of scale
        i0,i1,i2 = (
          int(idx-window-1),
          int(idx+window),
          int(idx+window+(tau*Ns/Ncyc))
        )
        w0 = arange(i0,i1)
        w1 = arange(i0,i2)
        fscl = slice(i0,i1) # Slice containing the training data
        a = polyfit(w0,unwrap(p)[fscl],1)
        scale0 = std(polyval(a,w0)-unwrap(p)[fscl]) # Scale
        scale = scale0
        
        # Now do M-estimator based regression with Tukey Biweight function
        c0 = 2.9
        
        
        
        # Useful
        
        alpha = 0.01
        b = a
        for i in xrange(10):
          M = array([
            [sum(dtukeyBiweight((unwrap(p)[fscl]-polyval(b,w0))/scale)*w0*w0),sum(dtukeyBiweight((unwrap(p)[fscl]-polyval(b,w0))/scale)*w0)],
            [sum(dtukeyBiweight((unwrap(p)[fscl]-polyval(b,w0))/scale)*w0),sum(dtukeyBiweight((unwrap(p)[fscl]-polyval(b,w0))/scale))]
          ])
          B = array([sum(tukeyBiweight((unwrap(p)[fscl]-polyval(b,w0))/scale)*w0),sum(tukeyBiweight((unwrap(p)[fscl]-polyval(b,w0))/scale))])
          b = b+alpha*dot(inv(M/scale),B)
          scale = std(polyval(b,w0)-unwrap(p)[fscl])
        
        '''
        tol = 1e-6
        ffac = 1e-2
        a0 = a[1]*array([(a[0]/a[1])*ffac,ffac])
        i = 0
        nscale = scale+tol+1
        while abs(scale-nscale)>tol:
          i+=1
          b = fmin(lambda bp: abs(sum(tukeyBiweight((unwrap(p)[fscl]-polyval(bp+a-a0,w0))/scale,c=c0))),a0,disp=False,ftol=1e-8)+a-a0
          scale = nscale
          nscale = std(polyval(b,w0)-unwrap(p)[fscl]) # Scale
          '''
        
        nscale = std(polyval(b,w0)-unwrap(p)[fscl]) # Scale
        print "Iterations:", i, "Method:", k
        print "Effective ignore fraction:", float(sum(abs((unwrap(p)[fscl]-polyval(b,w0))/scale)>c0))/len(p[fscl]), "Initial:", float(sum(abs((unwrap(p)[fscl]-polyval(a,w0))/scale)>c0))/len(p[fscl])
        print "Parameters (init, final):",b,a
        print "Scales (init, final, total change, final change):",scale0,nscale,scale0-nscale,scale-nscale
        print "Sum of psi functions (final, initial):", abs(sum(tukeyBiweight((unwrap(p)[fscl]-polyval(b,w0))/scale,c=c0))),abs(sum(tukeyBiweight((unwrap(p)[fscl]-polyval(a,w0))/scale,c=c0)))
        
        
        
        pp = polyval(b,w1)
        px = fls[k].val(pp).real.reshape(-1,5,2)
        pp2 = polyval(a,w1)
        px2 = fls[k].val(pp2).real.reshape(-1,5,2)
        slc = slice(i0,i2)
        resid = sum(sum((px-x[slc])**2,-1),-1) # To use normalised values divide by std(vstack(cxt)
        resid2 = sum(sum((px2-x[slc])**2,-1),-1) # To use normalised values divide by std(vstack(cxt)
        #resid = resid if sum(resid)<sum(resid2) else resid2
        
        semilogy(linspace(0,len(resid)/sr,len(resid)),resid,color=cols[k],alpha=0.6)
        perf[k].append(resid)
      fdtme.append(nidx-idx+2*window+1)

  #vlines((2*window+1)/sr,0,max(resid))

for k in phaseNames:
  perf[k] = array(perf[k])

figure()
for k in phaseNames:
  # MAKE RIBBONS AND MIDSTANCE LINES
  mu = mean(perf[k],0)
  sig = std(perf[k],0)/sqrt(11)
  tme = linspace(0,len(mu)/sr,len(mu))
  semilogy(tme,mu,color=cols[k],alpha=0.9)
  semilogy(tme,mu-sig,color=cols[k],alpha=0.5)
  semilogy(tme,mu+sig,color=cols[k],alpha=0.5)
  fill_between(tme,mu-sig,mu+sig,color=cols[k],alpha=0.2)
  xlim(min(tme),max(tme))
fill_betweenx(ylim(),tme[0],tme[2*window+1],color='y',alpha=0.1)
vlines(array(fdtme)/sr,*ylim(),alpha=0.3,color=cols['event'])
vlines((2*window+1)/sr,*ylim(),alpha=1.0,color=cols['event'])
xlabel("time (s)")
ylabel("mean square error (mm$^2$)")
savefig(glb.dog+"limbPositionError.png")

##############################################
# Bootstrap estimate of limit cycle and errors
##############################################
# Estimate limit cycle
Nbs = 1000
ls = dict()
fls = dict()
samp = list()

def buildLimCyc(cp,cx,theta,p=6):
  ls = list()
  for th0,th1 in zip(theta[:-1],theta[1:]):
    idx = logical_and(cp%(2*pi)>=th0,cp%(2*pi)<th1)
    ls.append(mean(cx[idx,:,:],0))
  ls = array(ls)
  # If we are missing entries interpolate  
  ls = nanInterpolation(ls.reshape(ls.shape[0],-1).T,method="linear").T.reshape(ls.shape)
  # Fourier expand
  fls = FourierSeries().fit(p,cp,cx.reshape(cp.shape[0],-1).T)
  return(ls,fls)  
  
theta = linspace(0,2*pi,251)
cx = vstack(cxt)
for k in phaseNames:
  ls[k] = list()
  fls[k] = list()
  cp = hstack(pht[k])
  for dummy in xrange(Nbs):
    if dummy%(100)==0:
      print "Doing:",k,dummy
    if len(samp)!=Nbs:
      samp.append(random.randint(0,len(cp),len(cp)))
    idx = samp[dummy]
    nls,nfls = buildLimCyc(cp[idx],cx[idx],theta)
    ls[k].append(nls)
    fls[k].append(nfls.val(theta[:-1]).real.reshape(-1,5,2).transpose(1,0,2))
  ls[k] = array(ls[k])
  fls[k] = array(fls[k])  
  
errD = dict()
figure()
errD['event'],bins,dump = hist(hstack(res['event']),25,color=cols['event'],alpha=0.4,normed=True)
for k in ['phaser','form']:
   errD[k],bins,dump = hist(hstack(res[k]),bins,color=cols[k],alpha=0.4,normed=True)

serrD = dict()
figure()
serrD['event'],sbins,dump = hist(hstack(res['event']),100,color=cols['event'],alpha=0.4,normed=True)
for k in ['phaser','form']:
   serrD[k],sbins,dump = hist(hstack(res[k]),sbins,color=cols[k],alpha=0.4,normed=True)


binLim = 80
hscale = 2.4
wscale = 2.2

AOIZ = ((-271.0,-140.0),70.0,58.0)

zorder = {'phaser':1, 'form':2, 'event':3}
limIdx = where(sbins>binLim)[0][0]
fig = figure(figsize=(14,9))
ax0 = subplot(1,2,1)
for k in phaseNames:
  for ix in fls[k].transpose(0,1,3,2):
    for iix in ix:
      ax0.plot(*iix,c=cols[k],lw=1,alpha=8.0/float(Nbs),zorder=zorder[k])
xlabel("fore-aft position (mm)")
ylabel("vertical position (mm)")
r = ax0.get_position()
ax0.add_patch(Rectangle(*AOIZ,facecolor='none',lw=2,edgecolor='k',zorder=4))

ax3 = fig.add_axes((r.x0+(wscale-1.0)*r.width/wscale,r.y0+(hscale-1)*r.height/hscale,r.width/wscale,r.height/hscale))

for k in phaseNames:
  for iix in fls[k].transpose(0,1,3,2)[:,3]:
    ax3.plot(*iix,c=cols[k],lw=3,alpha=8.0/float(Nbs),zorder=zorder[k])

ax3.set_xlim(AOIZ[0][0],AOIZ[0][0]+AOIZ[1])
ax3.set_ylim(AOIZ[0][1],AOIZ[0][1]+AOIZ[2])
ax3.set_xticks([])
ax3.set_yticks([])

ax1 = subplot(1,2,2)
for k in errD:
  ax1.plot(bins[:-1],errD[k],c=cols[k],marker=markers[k],lw=2)
semilogy(True)
xlabel("root mean square error (mm)")
ylabel("frequency density (1/mm)")
xlim([0,150])
ylim([0,max([max(errD[k]) for k in errD])*2.0])

r = ax1.get_position()
ax2 = fig.add_axes((r.x0+(wscale-1.0)*r.width/wscale,r.y0+(hscale-1)*r.height/hscale,r.width/wscale,r.height/hscale))
for k in serrD:
  ax2.plot(sbins[:limIdx],serrD[k][:limIdx],c=cols[k],marker=markers[k],lw=2)
ax2.semilogy(True)
y0,y1 = ax1.yaxis.axes.get_ybound()
x0,x1 = ax1.xaxis.axes.get_xbound()
ax2.set_ylim(y1/((y1/y0)**(1.0/hscale)),y1)
ax2.set_xlim(0,x1/wscale)

savefig(glb.dog+"doglimbmodelperformancesub.png")

# Single left figure
zorder = {'phaser':1, 'form':2, 'event':3}
limIdx = where(sbins>binLim)[0][0]
fig = figure(figsize=(14,14))
ax0 = subplot(1,1,1)
for k in phaseNames:
  for ix in fls[k].transpose(0,1,3,2):
    for iix in ix:
      ax0.plot(*iix,c=cols[k],lw=1,alpha=8.0/float(Nbs),zorder=zorder[k])
xlabel("fore-aft position (mm)")
ylabel("vertical position (mm)")
r = ax0.get_position()
ax0.add_patch(Rectangle(*AOIZ,facecolor='none',lw=2,edgecolor='k',zorder=4))
axis('equal')

ax3 = fig.add_axes((r.x0+(wscale-1.0)*r.width/wscale,r.y0+(hscale-1)*r.height/hscale,r.width/wscale,r.height/hscale))

for k in phaseNames:
  for iix in fls[k].transpose(0,1,3,2)[:,3]:
    ax3.plot(*iix,c=cols[k],lw=3,alpha=8.0/float(Nbs),zorder=zorder[k])

ax3.set_xlim(AOIZ[0][0],AOIZ[0][0]+AOIZ[1])
ax3.set_ylim(AOIZ[0][1],AOIZ[0][1]+AOIZ[2])
ax3.set_xticks([])
ax3.set_yticks([])
axis('equal')
savefig(glb.dog+"doglimbmodelperformancemin.png")

# Plot PRC
lcyc = vstack(median(ls['form'],0).transpose(1,2,0))
nlcyc = ((lcyc.T-zskmu)/zsksig).T
nlcyc = dot(U.T,nlcyc)
nlcyc = (nlcyc.T-zmed).T
pfrm = phi.form.getForm(nlcyc[:NpcF])


