'''
Extract the phases for each of 
the transitions for a dog and 
write these phases into a pickle.
@author Simon Wilshin
@email swilshin@rvc.ac.uk
@date Jul 2013
'''


from utility import splineDiff,getBeltSpeed
from options import glbOptions, getDirs

from sipylibs.qualisys import QualisysRawReader
from sipylibs import FiltFiltFilter,angularCentralTendancy

from formphase import formPhase
from formphase.formphaseutil import Phaser

from numpy import array,vstack,zeros_like,pi,linspace,unwrap,hstack,abs
from pylab import title,figure,plot,xlabel,ylabel,savefig,close,clf
from numpy.random import seed

from os.path import isfile
from os.path import join as osjoin
from os import listdir
from pickle import dump
from pickle import load
from sys import argv

seed(0) # Force consistent results

if len(argv)>=2 and argv[-1]!="--pylab":
  glb = glbOptions(dog=argv[-1])
elif '_dog' in locals():
  glb = glbOptions(dog=_dog)
else:
  print "WARNING: No dog specified, using default!"
  glb = glbOptions(None)


g3T = array([
                   [3.0/4.0, 1.0/2.0, 1.0/4.0],
                   [1.0/2.0, 1.0,     1.0/2.0],
                   [1.0/4.0, 1.0/2.0, 3.0/4.0],
                  ])

# Get files to process
dataDir,baseDir,outDir,intDir,logDir,plotDir = getDirs(glb.dog,glb.dataDir)
files = [f for f in listdir(intDir) if isfile(osjoin(intDir,f)) and f[-4:]==".tsv"]
trainFiles = [f for f in files if glb.reTraining.findall(f)!=[]]
testFiles = [f for f in files if not f in trainFiles]

# Grab the training data
trainfle = osjoin(outDir,"trainingData.pkl")
T,phaserTrain,beltSpeed,vdog,svdog,trainType = load(open(trainfle,'r'))

# Calculate derivatives of training data
x = dict()
v = dict()
for k in phaserTrain:
  x[k] = [vstack([ix.T,iv]) for ix,iv in zip(phaserTrain[k],svdog)]
  v[k] = [vstack([splineDiff(ix).T,zeros_like(iv)]) for ix,iv in zip(phaserTrain[k],beltSpeed)]
  
'''
Train the phasers we will need for each leg based on the 
training data
'''
phi = dict()
for k in ["ForeLeft","HindRight","HindLeft","ForeRight"]:
  if glb.usePhaser:
    phi[k]=Phaser([ix[:2] for ix in x[k] if ix.shape[1]>glb.minPhaserLen],ordP=20)
  else:
    phi[k]=formPhase(x[k],v[k],fourCor2=False,rbfPen=0.0,Nsb=100,ordPLim=20,minR=0.1)

# Now find offsets so that a trot 
# has correct phase offsets
trotResid = array([0,0,pi,pi])
trotPhases = dict()
for k in phi:
  figure()
  title(k)
  xlabel("frame (arb)")
  ylabel("residual phase (rad)")  
  for tpe,ix in zip(trainType,x[k]):
    if glb.usePhaser:
      theta = phi[k](ix[:2])[0]
    else:
      theta = phi[k](ix)
    slc = slice(glb.hilbertCut,-glb.hilbertCut)    
    if tpe=="Trot":
      if trotPhases.has_key(k):
        trotPhases[k].append(theta[slc])
      else:
        trotPhases[k] = [theta[slc]]
    plot(unwrap(theta[slc] - linspace(theta[slc][0],theta[slc][-1],theta[slc].size)))
  savefig(osjoin(plotDir,k+"trainPhase.svg"))

legs = ['ForeLeft', 'HindRight', 'HindLeft', 'ForeRight']
trotPhases = array([hstack(trotPhases[l]) for l in legs])
trotShift = array([0] + [angularCentralTendancy(ix-trotPhases[0]) for ix in trotPhases[1:]])-trotResid
trotShift = dict(zip(legs,trotShift))

print "Transforming testing data and getting belt speeds"
phaserTest = dict()
testBeltV = dict()
testvdog = dict()
testsvdog = dict()
sigs = dict()
for fle in testFiles:
  sigs[fle] = dict()
  Q = QualisysRawReader(osjoin(intDir,fle))
  print Q.NO_OF_MARKERS
  beltFilt = FiltFiltFilter(Q.FREQUENCY,glb.beltFilterFreq,3)
  phaserTest[fle] = T(Q)
  testBeltV[fle] = beltFilt(getBeltSpeed(Q,floorCond=glb.floorCond))[glb.cut:-glb.cut]
  testvdog[fle] = beltFilt(T.getvCoM(Q)[glb.cut:-glb.cut,0])
  testsvdog[fle] = abs(testvdog[fle]+testBeltV[fle])
  for k in phaserTest[fle]:
    sigs[fle][k] = vstack([phaserTest[fle][k].T,testsvdog[fle]])


print "Extracting phases"
phases = dict()
for fle in sigs:
  phases[fle] = dict()
  for k in phi:
    if glb.usePhaser:
      phases[fle][k] = phi[k](sigs[fle][k][:2])[0]-trotShift[k]
    else:
      phases[fle][k] = phi[k](sigs[fle][k])-trotShift[k]

# Now merge the phase differences
theta = dict()
for fle in sigs:
  theta[fle] = array([phases[fle][k]-phases[fle][q] for k,q in zip(legs[1:],legs[:-1])]).T%(2*pi)

# Plot some phase differences
for fle in sigs:
  figure()
  plot(theta[fle])
  title(fle)

close("all")
if False:
  # Get the user to classify the transition type
  g1 = dict()
  g2 = dict()
  figure()
  for k in theta:
    plot(theta[k]%(2*pi))
    g1[k] = raw_input("Gait L:")
    g2[k] = raw_input("Gait R:")
    clf()
  close("all")

  gTType = dict()
  for k in g1:
    gTType[k] = g1[k]+g2[k]

if True:
  fn = osjoin(outDir,"transitions.pkl")
  f = open(fn,'r')
  gTType,junk,junk,junk,junk,junk,junk,junk = load(f)
  f.close()

# Now save all the trajectories we have 
# extracted from each file.
fn = osjoin(outDir,"transitions.pkl")
f = open(fn,'w')
dump([gTType,sigs,phases,theta,testBeltV,testvdog,testsvdog,dict([(k,repr(phi[k])) for k in phi])],f,-1)
f.close()
