'''
Extract the training data for the gait trans 
project, write this data to file.
@author Simon Wilshin
@email swilshin@rvc.ac.uk
@date Jul 2013
'''
from options import glbOptions,getDirs
from utility import getBeltSpeed,GTTrans

from sipylibs.qualisys import QualisysRawReader
from sipylibs import FiltFiltFilter

from numpy import abs

from os.path import isfile
from os.path import join as osjoin
from os import listdir
from pickle import dump
from sys import argv

if len(argv)>=2 and argv[-1]!="--pylab":
  glb = glbOptions(dog=argv[-1])
else:
  print "WARNING: No dog specified, using default!"
  glb = glbOptions(None)

# Get files
print "Getting list of training files."
glb.dataDir,baseDir,outDir,intDir,logDir,plotDir = getDirs(glb.dog,glb.dataDir)
files = [f for f in listdir(intDir) if isfile(osjoin(intDir,f)) and f[-4:]==".tsv"]
trainFiles = [f for f in files if glb.reTraining.findall(f)!=[]]
trainType = ["Trot" if glb.reTraining.findall(f)[0]=="Trot" else "Walk" for f in trainFiles]

print "Assembling training data and computing belt speed."
sig = dict()
projections = dict()
psig = dict()

xcomRaw = dict()
beltSpeed = list()
vdog = list()
svdog = list()

print "Loading training data"
Qs = [ QualisysRawReader(osjoin(intDir,fle)) for fle in trainFiles]
print "Training transformation"
T = GTTrans(glb.NPC,glb.niceSignals,glb.lowFreq,glb.highFreq,glb.cut)
T.train(Qs)

print "Transforming training data"
phaserTrain = dict()
for Q in Qs:
  s = T(Q)
  for k in s:
    phaserTrain.setdefault(k,[]).append(s[k])

print "Calculating belt speed and dog speed"
for fle in trainFiles:
  Q = QualisysRawReader(osjoin(intDir,fle))
  sr = Q.FREQUENCY
  beltFilt = FiltFiltFilter(sr,glb.beltFilterFreq,3)
  beltSpeed.append(beltFilt(getBeltSpeed(Q,floorCond=glb.floorCond))[glb.cut:-glb.cut])
  vdog.append(beltFilt(T.getvCoM(Q)[glb.cut:-glb.cut,0]))
  svdog.append(abs(vdog[-1]+beltSpeed[-1]))

# Now save the projections and training data into 
# a single file.
print "Saving training data and transformation"
trainfle = open(osjoin(outDir,"trainingData.pkl"),'wb')
dump([T,phaserTrain,beltSpeed,vdog,svdog,trainType],trainfle)
trainfle.close()
